"""
Copyright 2026 The CVXPY Developers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import scipy.sparse as sp_sparse

# Type alias for bounds: (lower_bound, upper_bound)
Bounds = tuple[np.ndarray, np.ndarray]


def _ensure_dense(arr, shape: tuple[int, ...] | None = None):
    """Convert sparse matrix to dense numpy array and expand to shape if given.

    Parameters
    ----------
    arr : array-like
        The array to densify. May be sparse, a broadcast view, or a scalar.
    shape : tuple of int, optional
        Target shape to broadcast to. If None, returns the array as-is after
        densifying sparse matrices.

    Returns
    -------
    np.ndarray
        Dense array, optionally broadcast to the target shape.
    """
    if sp_sparse.issparse(arr):
        arr = np.asarray(arr.todense())
    else:
        arr = np.asarray(arr)

    if shape is not None and arr.shape != shape:
        # Use np.broadcast_to for memory-efficient expansion, then copy
        # to get a writable array (some solvers may need to modify bounds)
        arr = np.broadcast_to(arr, shape).copy()

    return arr


def _safe_maximum(a, b):
    """Element-wise maximum, handling sparse matrices.

    Uses scipy sparse's .maximum() when either operand is sparse,
    avoiding explicit densification of large sparse matrices.
    """
    if sp_sparse.issparse(a):
        return a.maximum(b)
    if sp_sparse.issparse(b):
        return b.maximum(a)
    return np.maximum(a, b)


def _safe_minimum(a, b):
    """Element-wise minimum, handling sparse matrices.

    Uses scipy sparse's .minimum() when either operand is sparse,
    avoiding explicit densification of large sparse matrices.
    """
    if sp_sparse.issparse(a):
        return a.minimum(b)
    if sp_sparse.issparse(b):
        return b.minimum(a)
    return np.minimum(a, b)


def _all_isinf(arr) -> bool:
    """Check if all values are inf, sparse-aware (O(nnz) for sparse)."""
    if sp_sparse.issparse(arr):
        # Structural zeros are 0, not inf. If there are any structural zeros
        # then not all values are inf.
        if arr.nnz < np.prod(arr.shape):
            return False
        return bool(np.all(np.isinf(arr.data)))
    return bool(np.all(np.isinf(arr)))


def _any_isnan(arr) -> bool:
    """Check if any values are NaN, sparse-aware (O(nnz) for sparse)."""
    if sp_sparse.issparse(arr):
        if arr.nnz == 0:
            return False
        return bool(np.any(np.isnan(arr.data)))
    return bool(np.any(np.isnan(arr)))


def _all_zero_or_inf(arr) -> bool:
    """Check if all values are 0 or inf, sparse-aware (O(nnz) for sparse)."""
    if sp_sparse.issparse(arr):
        # Structural zeros are 0, which satisfies the condition.
        # Only need to check explicitly stored values.
        if arr.nnz == 0:
            return True
        return bool(np.all((arr.data == 0) | np.isinf(arr.data)))
    return bool(np.all((arr == 0) | np.isinf(arr)))


def unbounded(shape: tuple[int, ...]) -> Bounds:
    """Return unbounded interval (-inf, inf) for given shape.

    Parameters
    ----------
    shape : tuple of ints
        The shape of the bounds arrays.

    Returns
    -------
    Bounds
        A tuple of (lower, upper) arrays filled with -inf and inf.
    """
    lower = np.full(shape, -np.inf)
    upper = np.full(shape, np.inf)
    return (lower, upper)


def uniform_bounds(shape: tuple[int, ...], lb: float, ub: float) -> Bounds:
    """Return uniform bounds as memory-efficient broadcast views.

    This creates read-only views that broadcast a scalar to the given shape
    without allocating memory for each element. Use this when all elements
    have the same bounds.

    Parameters
    ----------
    shape : tuple of ints
        The shape of the bounds arrays.
    lb : float
        The lower bound value (same for all elements).
    ub : float
        The upper bound value (same for all elements).

    Returns
    -------
    Bounds
        A tuple of broadcast views (lower, upper).
    """
    lb_scalar = np.array(lb)
    ub_scalar = np.array(ub)
    return (np.broadcast_to(lb_scalar, shape), np.broadcast_to(ub_scalar, shape))


def scalar_bounds(lb: float, ub: float) -> Bounds:
    """Return bounds for a scalar.

    Parameters
    ----------
    lb : float
        Lower bound.
    ub : float
        Upper bound.

    Returns
    -------
    Bounds
        A tuple of scalar arrays.
    """
    return (np.array(lb), np.array(ub))


def add_bounds(lb1: np.ndarray, ub1: np.ndarray,
               lb2: np.ndarray, ub2: np.ndarray) -> Bounds:
    """Bounds for elementwise addition: x + y.

    Parameters
    ----------
    lb1, ub1 : np.ndarray
        Bounds for the first operand.
    lb2, ub2 : np.ndarray
        Bounds for the second operand.

    Returns
    -------
    Bounds
        Bounds for the sum.
    """
    return (lb1 + lb2, ub1 + ub2)


def _sparse_sum(arr, axis, keepdims):
    """Sum a sparse array, handling keepdims which scipy doesn't support."""
    result = arr.sum(axis=axis)
    # scipy sparse .sum() returns np.matrix; convert to ndarray.
    result = np.asarray(result)
    if keepdims:
        result = np.expand_dims(result, axis=axis)
    else:
        result = result.squeeze()
    return result


def sum_bounds(lb: np.ndarray, ub: np.ndarray,
               axis: int | tuple[int, ...] | None = None,
               keepdims: bool = False) -> Bounds:
    """Bounds for sum reduction.

    Parameters
    ----------
    lb, ub : array-like
        Bounds for the expression being summed (may be sparse).
    axis : None or int or tuple of ints
        Axis or axes along which to sum.
    keepdims : bool
        Whether to keep the reduced dimensions.

    Returns
    -------
    Bounds
        Bounds for the sum.
    """
    if sp_sparse.issparse(lb):
        new_lb = _sparse_sum(lb, axis, keepdims)
    else:
        new_lb = np.sum(lb, axis=axis, keepdims=keepdims)
    if sp_sparse.issparse(ub):
        new_ub = _sparse_sum(ub, axis, keepdims)
    else:
        new_ub = np.sum(ub, axis=axis, keepdims=keepdims)
    return (new_lb, new_ub)


def neg_bounds(lb, ub) -> Bounds:
    """Bounds for negation: -x.

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the expression.

    Returns
    -------
    Bounds
        Bounds for the negation.
    """
    return (-ub, -lb)


def mul_bounds(lb1, ub1, lb2, ub2) -> Bounds:
    """Bounds for elementwise multiplication: x * y.

    Uses interval arithmetic: the result is [min(products), max(products)]
    where products are all combinations of endpoints.

    Handles sparse matrices without densifying them unnecessarily.

    Parameters
    ----------
    lb1, ub1 : array-like
        Bounds for the first operand (may be sparse).
    lb2, ub2 : array-like
        Bounds for the second operand (may be sparse).

    Returns
    -------
    Bounds
        Bounds for the product.
    """
    # All four products of interval endpoints.
    # Replace NaN with 0: the only source of NaN here is 0 * inf,
    # and in interval arithmetic 0 * anything = 0.
    def _mul(a, b):
        with np.errstate(invalid="ignore"):
            p = a * b
        if sp_sparse.issparse(p):
            p.data = np.where(np.isnan(p.data), 0.0, p.data)
            return p
        return np.where(np.isnan(p), 0.0, p)

    p1 = _mul(lb1, lb2)
    p2 = _mul(lb1, ub2)
    p3 = _mul(ub1, lb2)
    p4 = _mul(ub1, ub2)

    # Use iterative min/max when sparse matrices are involved
    if any(sp_sparse.issparse(p) for p in [p1, p2, p3, p4]):
        new_lb = p1
        new_ub = p1
        for p in [p2, p3, p4]:
            new_lb = _safe_minimum(new_lb, p)
            new_ub = _safe_maximum(new_ub, p)
        return (new_lb, new_ub)

    # Dense path: stack and take min/max
    products = np.stack([p1, p2, p3, p4], axis=0)
    new_lb = np.min(products, axis=0)
    new_ub = np.max(products, axis=0)
    return (new_lb, new_ub)


def div_bounds(lb1, ub1, lb2, ub2) -> Bounds:
    """Bounds for elementwise division: x / y.

    Note: If the divisor interval contains zero, returns unbounded.

    Parameters
    ----------
    lb1, ub1 : array-like
        Bounds for the numerator (may be sparse).
    lb2, ub2 : array-like
        Bounds for the divisor (may be sparse).

    Returns
    -------
    Bounds
        Bounds for the quotient.
    """
    # Reciprocal of a sparse matrix is inherently dense (1/0 for structural zeros),
    # so convert sparse divisor bounds to dense.
    lb2 = _ensure_dense(lb2)
    ub2 = _ensure_dense(ub2)

    # Check for division by interval containing zero
    contains_zero = (lb2 <= 0) & (ub2 >= 0)

    # Compute 1/[lb2, ub2]
    # When lb2 and ub2 have the same sign, reciprocal reverses the interval
    inv_lb = np.where(contains_zero, -np.inf, 1.0 / ub2)
    inv_ub = np.where(contains_zero, np.inf, 1.0 / lb2)

    # Now multiply by numerator bounds
    return mul_bounds(lb1, ub1, inv_lb, inv_ub)


def scale_bounds(lb: np.ndarray, ub: np.ndarray, c: float) -> Bounds:
    """Bounds for scalar multiplication: c * x.

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the expression.
    c : float
        The scalar multiplier.

    Returns
    -------
    Bounds
        Bounds for the scaled expression.
    """
    if c >= 0:
        return (c * lb, c * ub)
    else:
        return (c * ub, c * lb)


def abs_bounds(lb: np.ndarray, ub: np.ndarray) -> Bounds:
    """Bounds for elementwise absolute value: |x|.

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the expression.

    Returns
    -------
    Bounds
        Bounds for the absolute value.
    """
    # When interval spans zero, lower bound is 0
    # When interval is entirely positive, |x| = x
    # When interval is entirely negative, |x| = -x
    spans_zero = (lb <= 0) & (ub >= 0)
    entirely_positive = lb >= 0
    entirely_negative = ub <= 0

    new_lb = np.where(spans_zero, 0.0,
                      np.where(entirely_positive, lb, -ub))
    new_ub = np.where(entirely_positive, ub,
                      np.where(entirely_negative, -lb,
                               np.maximum(-lb, ub)))
    return (new_lb, new_ub)


def maximum_bounds(bounds_list: list[Bounds]) -> Bounds:
    """Bounds for elementwise maximum: max(x1, x2, ...).

    Parameters
    ----------
    bounds_list : list of Bounds
        List of (lb, ub) tuples for each argument.

    Returns
    -------
    Bounds
        Bounds for the maximum.
    """
    # Use reduce with np.maximum to handle broadcasting between different shapes
    lb_result = bounds_list[0][0]
    ub_result = bounds_list[0][1]
    for lb, ub in bounds_list[1:]:
        lb_result = np.maximum(lb_result, lb)
        ub_result = np.maximum(ub_result, ub)
    return (lb_result, ub_result)


def minimum_bounds(bounds_list: list[Bounds]) -> Bounds:
    """Bounds for elementwise minimum: min(x1, x2, ...).

    Parameters
    ----------
    bounds_list : list of Bounds
        List of (lb, ub) tuples for each argument.

    Returns
    -------
    Bounds
        Bounds for the minimum.
    """
    # Use reduce with np.minimum to handle broadcasting between different shapes
    lb_result = bounds_list[0][0]
    ub_result = bounds_list[0][1]
    for lb, ub in bounds_list[1:]:
        lb_result = np.minimum(lb_result, lb)
        ub_result = np.minimum(ub_result, ub)
    return (lb_result, ub_result)


def max_reduction_bounds(lb: np.ndarray, ub: np.ndarray,
                         axis: int | tuple[int, ...] | None = None,
                         keepdims: bool = False) -> Bounds:
    """Bounds for max reduction: max(x, axis=axis).

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the expression.
    axis : None or int or tuple of ints
        Axis or axes along which to take max.
    keepdims : bool
        Whether to keep the reduced dimensions.

    Returns
    -------
    Bounds
        Bounds for the max.
    """
    new_lb = np.max(lb, axis=axis, keepdims=keepdims)
    new_ub = np.max(ub, axis=axis, keepdims=keepdims)
    return (new_lb, new_ub)


def min_reduction_bounds(lb: np.ndarray, ub: np.ndarray,
                         axis: int | tuple[int, ...] | None = None,
                         keepdims: bool = False) -> Bounds:
    """Bounds for min reduction: min(x, axis=axis).

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the expression.
    axis : None or int or tuple of ints
        Axis or axes along which to take min.
    keepdims : bool
        Whether to keep the reduced dimensions.

    Returns
    -------
    Bounds
        Bounds for the min.
    """
    new_lb = np.min(lb, axis=axis, keepdims=keepdims)
    new_ub = np.min(ub, axis=axis, keepdims=keepdims)
    return (new_lb, new_ub)


def power_bounds(lb: np.ndarray, ub: np.ndarray, p: float) -> Bounds:
    """Bounds for elementwise power: x^p.

    Handles different cases based on p:
    - p > 0: behavior depends on whether p is even/odd and sign of interval
    - p < 0: requires positive interval
    - p = 0: constant 1

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the base expression.
    p : float
        The power.

    Returns
    -------
    Bounds
        Bounds for the power.
    """
    if p == 0:
        return (np.ones_like(lb), np.ones_like(ub))

    if p > 0:
        if p == int(p) and int(p) % 2 == 0:
            # Even integer power: x^(2k)
            # If interval entirely positive: [lb^p, ub^p]
            # If interval entirely negative: [ub^p, lb^p]
            # If interval spans zero: [0, max(lb^p, ub^p)]
            spans_zero = (lb <= 0) & (ub >= 0)
            entirely_positive = lb >= 0

            lb_power = np.abs(lb) ** p
            ub_power = np.abs(ub) ** p

            new_lb = np.where(spans_zero, 0.0,
                              np.where(entirely_positive, lb_power, ub_power))
            new_ub = np.where(spans_zero, np.maximum(lb_power, ub_power),
                              np.where(entirely_positive, ub_power, lb_power))
            return (new_lb, new_ub)
        else:
            # Odd integer power or non-integer positive power
            # For odd powers, x^p is monotonic: [lb^p, ub^p]
            # For non-integer powers, typically only defined for x >= 0
            # We'll handle the general case conservatively
            if p == int(p):
                # Odd integer power: monotonic
                return (lb ** p, ub ** p)
            else:
                # Non-integer power: requires x >= 0 for real result
                # If lb < 0, we have undefined behavior, return unbounded for those
                valid = lb >= 0
                new_lb = np.where(valid, lb ** p, -np.inf)
                new_ub = np.where(valid, ub ** p, np.inf)
                return (new_lb, new_ub)
    else:
        # Negative power: x^(-|p|) = 1/x^|p|
        # Only defined for x != 0
        # For positive intervals: [ub^p, lb^p] (reverses order)
        # For negative intervals with odd |p|: [ub^p, lb^p]
        # Intervals spanning zero: unbounded
        spans_zero = (lb <= 0) & (ub >= 0)
        entirely_positive = lb > 0

        if p == int(p) and int(-p) % 2 == 1:
            # Negative odd power: monotonically decreasing
            new_lb = np.where(spans_zero, -np.inf, np.minimum(lb ** p, ub ** p))
            new_ub = np.where(spans_zero, np.inf, np.maximum(lb ** p, ub ** p))
        else:
            # Negative even power or non-integer: only for positive x
            new_lb = np.where(entirely_positive, ub ** p, -np.inf)
            new_ub = np.where(entirely_positive, lb ** p, np.inf)

        return (new_lb, new_ub)


def exp_bounds(lb: np.ndarray, ub: np.ndarray) -> Bounds:
    """Bounds for elementwise exponential: exp(x).

    exp is monotonically increasing, so bounds are [exp(lb), exp(ub)].

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the argument.

    Returns
    -------
    Bounds
        Bounds for the exponential.
    """
    return (np.exp(lb), np.exp(ub))


def log_bounds(lb: np.ndarray, ub: np.ndarray) -> Bounds:
    """Bounds for elementwise natural log: log(x).

    log is monotonically increasing on (0, inf), undefined for x <= 0.

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the argument.

    Returns
    -------
    Bounds
        Bounds for the logarithm.
    """
    # log is only defined for positive arguments
    valid = lb > 0
    new_lb = np.where(valid, np.log(lb), -np.inf)
    new_ub = np.where(ub > 0, np.log(ub), np.inf)
    return (new_lb, new_ub)


def sqrt_bounds(lb: np.ndarray, ub: np.ndarray) -> Bounds:
    """Bounds for elementwise square root: sqrt(x).

    sqrt is monotonically increasing on [0, inf).

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the argument (must be >= 0).

    Returns
    -------
    Bounds
        Bounds for the square root.
    """
    # sqrt is only defined for non-negative arguments
    new_lb = np.where(lb >= 0, np.sqrt(lb), -np.inf)
    new_ub = np.where(ub >= 0, np.sqrt(ub), np.inf)
    return (new_lb, new_ub)


def norm1_bounds(lb: np.ndarray, ub: np.ndarray,
                 axis: int | tuple[int, ...] | None = None,
                 keepdims: bool = False) -> Bounds:
    """Bounds for 1-norm: sum(|x|).

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the expression.
    axis : None or int or tuple of ints
        Axis along which to compute the norm.
    keepdims : bool
        Whether to keep the reduced dimensions.

    Returns
    -------
    Bounds
        Bounds for the 1-norm.
    """
    abs_lb, abs_ub = abs_bounds(lb, ub)
    return sum_bounds(abs_lb, abs_ub, axis=axis, keepdims=keepdims)


def norm_inf_bounds(lb: np.ndarray, ub: np.ndarray,
                    axis: int | tuple[int, ...] | None = None,
                    keepdims: bool = False) -> Bounds:
    """Bounds for infinity-norm: max(|x|).

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds for the expression.
    axis : None or int or tuple of ints
        Axis along which to compute the norm.
    keepdims : bool
        Whether to keep the reduced dimensions.

    Returns
    -------
    Bounds
        Bounds for the infinity norm.
    """
    abs_lb, abs_ub = abs_bounds(lb, ub)
    return max_reduction_bounds(abs_lb, abs_ub, axis=axis, keepdims=keepdims)


def broadcast_bounds(lb, ub, target_shape: tuple[int, ...]) -> Bounds:
    """Broadcast bounds to a target shape.

    Handles sparse matrices: if a sparse bound already has the target shape,
    it is kept as-is; otherwise it is converted to dense for broadcasting.

    Parameters
    ----------
    lb, ub : array-like
        Bounds to broadcast (may be sparse).
    target_shape : tuple of ints
        Target shape.

    Returns
    -------
    Bounds
        Broadcasted bounds.
    """
    def _broadcast_one(arr, shape):
        if sp_sparse.issparse(arr):
            if arr.shape == shape:
                return arr
            # Can't broadcast sparse to a different shape; convert to dense.
            return np.broadcast_to(arr.toarray(), shape)
        return np.broadcast_to(arr, shape)

    return (_broadcast_one(lb, target_shape), _broadcast_one(ub, target_shape))


def reshape_bounds(lb: np.ndarray, ub: np.ndarray,
                   new_shape: tuple[int, ...],
                   order: str = 'F') -> Bounds:
    """Reshape bounds to a new shape.

    Parameters
    ----------
    lb, ub : array-like
        Bounds to reshape (may be sparse).
    new_shape : tuple of ints
        New shape.
    order : str
        'C' for row-major (C-style) or 'F' for column-major (Fortran-style).

    Returns
    -------
    Bounds
        Reshaped bounds.
    """
    return (lb.reshape(new_shape, order=order), ub.reshape(new_shape, order=order))


def transpose_bounds(lb: np.ndarray, ub: np.ndarray,
                     axes: tuple[int, ...] | None = None) -> Bounds:
    """Transpose bounds.

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds to transpose.
    axes : tuple of ints or None
        Permutation of axes.

    Returns
    -------
    Bounds
        Transposed bounds.
    """
    return (np.transpose(lb, axes), np.transpose(ub, axes))


def index_bounds(lb: np.ndarray, ub: np.ndarray, key) -> Bounds:
    """Index into bounds.

    Parameters
    ----------
    lb, ub : np.ndarray
        Bounds to index.
    key : index
        Index/slice to apply.

    Returns
    -------
    Bounds
        Indexed bounds.
    """
    return (lb[key], ub[key])


def matmul_bounds(lb1, ub1, lb2, ub2) -> Bounds:
    """Bounds for matrix multiplication: x @ y.

    When one operand is a point (lb == ub, i.e. a constant), uses the
    exact split formula: A_pos @ lb2 + A_neg @ ub2. When both operands
    are intervals, returns unbounded since there is no efficient exact
    formula.

    Handles sparse matrices efficiently using scipy sparse operations
    to avoid densifying large sparse matrices.

    Parameters
    ----------
    lb1, ub1 : array-like
        Bounds for the first matrix (may be sparse).
    lb2, ub2 : array-like
        Bounds for the second matrix (may be sparse).

    Returns
    -------
    Bounds
        Bounds for the matrix product.
    """
    # The split formula (A_pos @ B_l + A_neg @ B_u) is only valid when
    # one operand is a known point (lb == ub). When both are intervals
    # the per-element minimum of a*b requires element-wise operations
    # that can't be expressed as matrix products, so we return unbounded.
    def _is_point(lb, ub):
        if sp_sparse.issparse(lb) and sp_sparse.issparse(ub):
            diff = lb - ub
            return diff.nnz == 0
        return np.array_equal(lb, ub)

    lhs_point = _is_point(lb1, ub1)
    rhs_point = _is_point(lb2, ub2)

    if lhs_point:
        # A is a constant matrix. C = A @ B where B in [lb2, ub2].
        # C_ij = sum_k A_ik * B_kj; min over B gives:
        #   A_pos @ lb2 + A_neg @ ub2
        a_pos = _safe_maximum(lb1, 0)
        a_neg = _safe_minimum(lb1, 0)
        new_lb = a_pos @ lb2 + a_neg @ ub2
        new_ub = a_pos @ ub2 + a_neg @ lb2
        return (new_lb, new_ub)

    if rhs_point:
        # B is a constant matrix. C = A @ B where A in [lb1, ub1].
        # C_ij = sum_k A_ik * B_kj; min over A gives:
        #   lb1 @ B_pos + ub1 @ B_neg  (for lower bound)
        b_pos = _safe_maximum(lb2, 0)
        b_neg = _safe_minimum(lb2, 0)
        new_lb = lb1 @ b_pos + ub1 @ b_neg
        new_ub = ub1 @ b_pos + lb1 @ b_neg
        return (new_lb, new_ub)

    # Both operands are intervals — no efficient exact formula.
    shape = (np.shape(lb1)[0],) if np.ndim(lb2) == 1 else (
        np.shape(lb1)[0], np.shape(lb2)[1])
    return unbounded(shape)


def get_expr_bounds_if_supported(expr, solver_context) -> list | None:
    """Get bounds from expression for use on auxiliary variables.

    Returns a [lb, ub] list suitable for passing to Variable(bounds=...),
    or None if bounds should not be set. Returns None when:
    - The solver does not support bounded variables.
    - The expression bounds are entirely unbounded (-inf, inf).
    - The bounds contain NaN values.
    - The bounds only encode sign information that is already captured
      by the expression's sign attributes (avoids redundant constraints).

    Parameters
    ----------
    expr : Expression
        The expression whose bounds to compute.
    solver_context : SolverInfo or None
        Solver context; if None or the solver does not support bounds,
        returns None immediately.

    Returns
    -------
    list or None
        [lb, ub] arrays, or None.
    """
    if solver_context is None or not solver_context.solver_supports_bounds:
        return None
    lb, ub = expr.get_bounds()
    # Check if bounds are finite and worth using.
    # Use sparse-aware checks to avoid densifying large sparse matrices.
    if _all_isinf(lb) and _all_isinf(ub):
        return None
    # Check for NaN values which are not valid bounds
    if _any_isnan(lb) or _any_isnan(ub):
        return None
    # Check if bounds only match sign info (avoid redundant constraints)
    # If lb is all 0 or -inf, and ub is all inf, and expr is nonneg, skip
    lb_trivial = _all_zero_or_inf(lb)
    ub_trivial = _all_isinf(ub)
    if lb_trivial and ub_trivial and expr.is_nonneg():
        return None
    # If ub is all 0 or inf, and lb is all -inf, and expr is nonpos, skip
    ub_trivial_nonpos = _all_zero_or_inf(ub)
    lb_trivial_nonpos = _all_isinf(lb)
    if lb_trivial_nonpos and ub_trivial_nonpos and expr.is_nonpos():
        return None
    # Ensure dense and expand to full shape for solver consumption.
    # This handles broadcast views and sparse matrices.
    return [_ensure_dense(lb, expr.shape), _ensure_dense(ub, expr.shape)]


def coords_equal(coords1, coords2) -> bool:
    """Check if two coordinate tuples represent the same sparsity pattern.

    Parameters
    ----------
    coords1 : tuple of array-like
        First coordinate tuple (e.g., (rows, cols) for 2D sparse).
    coords2 : tuple of array-like
        Second coordinate tuple.

    Returns
    -------
    bool
        True if the coordinate tuples are equal.
    """
    if len(coords1) != len(coords2):
        return False
    return all(np.array_equal(c1, c2) for c1, c2 in zip(coords1, coords2))


def refine_bounds_from_sign(lb, ub,
                            is_nonneg: bool, is_nonpos: bool) -> Bounds:
    """Refine bounds based on sign information.

    Parameters
    ----------
    lb, ub : array-like
        Current bounds (may be sparse).
    is_nonneg : bool
        Whether the expression is known to be non-negative.
    is_nonpos : bool
        Whether the expression is known to be non-positive.

    Returns
    -------
    Bounds
        Refined bounds.
    """
    if is_nonneg:
        lb = _safe_maximum(lb, 0)
    if is_nonpos:
        ub = _safe_minimum(ub, 0)
    return (lb, ub)



"""
Copyright 2025 CVXPY developers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from cvxpy.atoms.elementwise.rel_entr import rel_entr
from cvxpy.reductions.dnlp2smooth.canonicalizers.rel_entr_canon import rel_entr_canon


def kl_div_canon(expr, args):
    _rel_entr = rel_entr(args[0], args[1])
    rel_entr_expr, constr = rel_entr_canon(_rel_entr, _rel_entr.args)
    return rel_entr_expr - args[0] + args[1] , constr




"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import numpy.linalg as LA
import pytest

import cvxpy as cp
from cvxpy.reductions.solvers.defines import INSTALLED_SOLVERS
from cvxpy.tests.nlp_tests.derivative_checker import DerivativeChecker


@pytest.mark.skipif('IPOPT' not in INSTALLED_SOLVERS, reason='IPOPT is not installed.')
class TestEntropy():

    # convex problem, standard entropy f(x) = - x log x.
    def test_entropy_one(self):
       np.random.seed(0)
       n = 100
       q = cp.Variable(n, nonneg=True)
       A = np.random.rand(n, n)
       obj = cp.sum(cp.entr(A @ q))
       constraints = [cp.sum(q) == 1]
       problem = cp.Problem(cp.Maximize(obj), constraints)
       problem.solve(solver=cp.IPOPT, nlp=True, verbose=True, derivative_test='none')
       q_opt_nlp = q.value
       problem.solve(solver=cp.CLARABEL, verbose=True)
       q_opt_clarabel = q.value
       assert(LA.norm(q_opt_nlp - q_opt_clarabel) <= 1e-4)

       checker = DerivativeChecker(problem)
       checker.run_and_assert()

    # nonconvex problem, compute minimum entropy distribution
    # over simplex (the analytical solution is any of the vertices)
    def test_entropy_two(self):
        np.random.seed(0)
        n = 10
        q = cp.Variable((n, ), nonneg=True)
        q.value = np.random.rand(n)
        q.value = q.value / np.sum(q.value)
        obj = cp.sum(cp.entr(q))
        constraints = [cp.sum(q) == 1]
        problem = cp.Problem(cp.Minimize(obj), constraints)
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True,
                      hessian_approximation='limited-memory')
        q_opt_nlp = q.value
        assert(np.sum(q_opt_nlp > 1e-8) == 1)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    # convex formulation, relative entropy f(x, y) = x log (x / y)
    def test_rel_entropy_one(self):
        np.random.seed(0)
        n = 40
        p = np.random.rand(n, )
        p = p / np.sum(p)
        q = cp.Variable(n, nonneg=True)
        A = np.random.rand(n, n)
        obj = cp.sum(cp.rel_entr(A @ q, p))
        constraints = [cp.sum(q) == 1]
        problem = cp.Problem(cp.Minimize(obj), constraints)
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True,  derivative_test='none')
        q_opt_nlp = q.value
        problem.solve(solver=cp.CLARABEL, verbose=True)
        q_opt_clarabel = q.value
        assert(LA.norm(q_opt_nlp - q_opt_clarabel) <= 1e-4)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_rel_entropy_one_switched_arguments(self):
        np.random.seed(0)
        n = 40
        p = np.random.rand(n, )
        p = p / np.sum(p)
        q = cp.Variable(n, nonneg=True)
        A = np.random.rand(n, n)
        obj = cp.sum(cp.rel_entr(p, A @ q))
        constraints = [cp.sum(q) == 1]
        problem = cp.Problem(cp.Minimize(obj), constraints)
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True,  derivative_test='none')
        q_opt_nlp = q.value
        problem.solve(solver=cp.CLARABEL, verbose=True)
        q_opt_clarabel = q.value
        assert(LA.norm(q_opt_nlp - q_opt_clarabel) <= 1e-4)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_KL_one(self):
        np.random.seed(0)
        n = 40
        p = np.random.rand(n, )
        p = p / np.sum(p)
        q = cp.Variable(n, nonneg=True)
        A = np.random.rand(n, n)
        obj = cp.sum(cp.kl_div(A @ q, p))
        constraints = [cp.sum(q) == 1]
        problem = cp.Problem(cp.Minimize(obj), constraints)
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True,  derivative_test='none')
        q_opt_nlp = q.value
        problem.solve(solver=cp.CLARABEL, verbose=True)
        q_opt_clarabel = q.value
        assert(LA.norm(q_opt_nlp - q_opt_clarabel) <= 1e-4)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_KL_two(self):
        np.random.seed(0)
        n = 40
        p = np.random.rand(n, )
        p = p / np.sum(p)
        q = cp.Variable(n, nonneg=True)
        A = np.random.rand(n, n)
        obj = cp.sum(cp.kl_div(p, A @ q))
        constraints = [cp.sum(q) == 1]
        problem = cp.Problem(cp.Minimize(obj), constraints)
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True,  derivative_test='none')
        q_opt_nlp = q.value
        problem.solve(solver=cp.CLARABEL, verbose=True)
        q_opt_clarabel = q.value
        assert(LA.norm(q_opt_nlp - q_opt_clarabel) <= 1e-4)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    # nonnegative matrix factorization with KL objective (nonconvex)
    def test_KL_three_graph_form(self):
        np.random.seed(0)
        n, m, k = 40, 20, 4
        X_true = np.random.rand(n, k)
        Y_true = np.random.rand(k, m)
        A = X_true @ Y_true
        A = np.clip(A, 0, None)
        X = cp.Variable((n, k), bounds=[0, None])
        Y = cp.Variable((k, m), bounds=[0, None])
        # without random initialization we converge to a very structured
        # point that is not the global minimizer
        X.value = np.random.rand(n, k)
        Y.value = np.random.rand(k, m)

        # graph form
        T = cp.Variable((n, m))
        obj = cp.sum(cp.kl_div(A, T))
        constraints = [T == X @ Y]
        problem = cp.Problem(cp.Minimize(obj), constraints)
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False, derivative_test='none')
        assert(obj.value <= 1e-10)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    # nonnegative matrix factorization with KL objective (nonconvex)
    def test_KL_three_not_graph_form(self):
        np.random.seed(0)
        n, m, k = 40, 20, 4
        X_true = np.random.rand(n, k)
        Y_true = np.random.rand(k, m)
        A = X_true @ Y_true
        A = np.clip(A, 0, None)
        X = cp.Variable((n, k), bounds=[0, None])
        Y = cp.Variable((k, m), bounds=[0, None])
        X.value = np.random.rand(n, k)
        Y.value = np.random.rand(k, m)
        obj = cp.sum(cp.kl_div(A, X @ Y))
        problem = cp.Problem(cp.Minimize(obj))
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False, derivative_test='none')
        assert(obj.value <= 1e-10)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import numpy as np
import pytest

import cvxpy as cp
from cvxpy.reductions.solvers.defines import INSTALLED_SOLVERS
from cvxpy.tests.nlp_tests.derivative_checker import DerivativeChecker
from cvxpy.tests.test_conic_solvers import is_knitro_available

# Always parametrize all solvers, skip at runtime if not available
NLP_SOLVERS = [
    pytest.param('IPOPT', marks=pytest.mark.skipif(
        'IPOPT' not in INSTALLED_SOLVERS, reason='IPOPT is not installed.')),
    pytest.param('KNITRO', marks=pytest.mark.skipif(
        not is_knitro_available(), reason='KNITRO is not installed or license not available.')),
    pytest.param('UNO', marks=pytest.mark.skipif(
        'UNO' not in INSTALLED_SOLVERS, reason='UNO is not installed.')),
    pytest.param('COPT', marks=pytest.mark.skipif(
        'COPT' not in INSTALLED_SOLVERS, reason='COPT is not installed.')),
]


@pytest.mark.parametrize("solver", NLP_SOLVERS)
class TestNLPExamples:
    """
    Nonlinear test problems taken from the IPOPT documentation and
    the Julia documentation: https://jump.dev/JuMP.jl/stable/tutorials/nonlinear/simple_examples/.
    """

    def test_hs071(self, solver):
        x = cp.Variable(4, bounds=[0, 6])
        x.value = np.array([1.0, 5.0, 5.0, 1.0])
        objective = cp.Minimize(x[0]*x[3]*(x[0] + x[1] + x[2]) + x[2])

        constraints = [
            x[0]*x[1]*x[2]*x[3] >= 25,
            cp.sum(cp.square(x)) == 40,
        ]
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(x.value, np.array([0.75450865, 4.63936861, 3.78856881, 1.88513184]))

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_mle(self, solver):
        n = 1000
        np.random.seed(1234)
        data = np.random.randn(n)

        mu = cp.Variable((1,), name="mu")
        mu.value = np.array([0.0])
        sigma = cp.Variable((1,), name="sigma")
        sigma.value = np.array([1.0])

        constraints = [mu == sigma**2]
        log_likelihood = (
            (n / 2) * cp.log(1 / (2 * np.pi * (sigma)**2))
            - cp.sum(cp.square(data-mu)) / (2 * (sigma)**2)
        )

        objective = cp.Maximize(log_likelihood)
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(sigma.value, 0.77079388)
        assert np.allclose(mu.value, 0.59412321)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_portfolio_opt(self, solver):
        # data taken from https://jump.dev/JuMP.jl/stable/tutorials/nonlinear/portfolio/
        # r and Q are pre-computed from historical data of 3 assets
        r = np.array([0.026002150277777, 0.008101316405671, 0.073715909491990])
        Q = np.array([
            [0.018641039983891, 0.003598532927677, 0.001309759253660],
            [0.003598532927677, 0.006436938322676, 0.004887265158407],
            [0.001309759253660, 0.004887265158407, 0.068682765454814],
        ])
        x = cp.Variable(3)
        x.value = np.array([10.0, 10.0, 10.0])
        variance = cp.quad_form(x, Q)
        expected_return = r @ x
        problem = cp.Problem(
            cp.Minimize(variance),
            [
                cp.sum(x) <= 1000,
                expected_return >= 50,
                x >= 0
            ]
        )
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        # Second element can be slightly negative due to numerical tolerance
        assert np.allclose(x.value, np.array([4.97045504e+02, 0.0, 5.02954496e+02]), atol=1e-4)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_portfolio_opt_sum_multiply(self, solver):
        # data taken from https://jump.dev/JuMP.jl/stable/tutorials/nonlinear/portfolio/
        # r and Q are pre-computed from historical data of 3 assets
        r = np.array([0.026002150277777, 0.008101316405671, 0.073715909491990])
        Q = np.array([
            [0.018641039983891, 0.003598532927677, 0.001309759253660],
            [0.003598532927677, 0.006436938322676, 0.004887265158407],
            [0.001309759253660, 0.004887265158407, 0.068682765454814],
        ])
        x = cp.Variable(3)
        x.value = np.array([10.0, 10.0, 10.0])
        variance = cp.quad_form(x, Q)
        expected_return = cp.sum(cp.multiply(r, x))
        problem = cp.Problem(
            cp.Minimize(variance),
            [
                cp.sum(x) <= 1000,
                expected_return >= 50,
                x >= 0
            ]
        )
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        # Second element can be slightly negative due to numerical tolerance
        assert np.allclose(x.value, np.array([4.97045504e+02, 0.0, 5.02954496e+02]), atol=1e-4)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    @pytest.mark.skipif('UNO' in INSTALLED_SOLVERS, reason='UNO has an algorithmic error')
    def test_rosenbrock(self, solver):
        x = cp.Variable(2, name='x')
        objective = cp.Minimize((1 - x[0])**2 + 100 * (x[1] - x[0]**2)**2)
        problem = cp.Problem(objective, [])
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(x.value, np.array([1.0, 1.0]))
        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_qcp(self, solver):
        x = cp.Variable(1)
        y = cp.Variable(1, bounds=[0, np.inf])
        z = cp.Variable(1, bounds=[0, np.inf])

        objective = cp.Maximize(x)

        constraints = [
            x + y + z == 1,
            x**2 + y**2 - z**2 <= 0,
            x**2 - cp.multiply(y, z) <= 0
        ]
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(x.value, np.array([0.32699284]))
        assert np.allclose(y.value, np.array([0.25706586]))
        assert np.allclose(z.value, np.array([0.4159413]))

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_analytic_polytope_center(self, solver):
        # Generate random data
        np.random.seed(0)
        m, n = 50, 4
        b = np.ones(m)
        rand = np.random.randn(m - 2*n, n)
        A = np.vstack((rand, np.eye(n), np.eye(n) * -1))

        # Define the variable
        x = cp.Variable(n)
        # set initial value for x
        objective = cp.Minimize(-cp.sum(cp.log(b - A @ x)))
        problem = cp.Problem(objective, [])
        # Solve the problem
        problem.solve(solver=solver, nlp=True)

        assert problem.status == cp.OPTIMAL

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_analytic_polytope_center_x_column_vector(self, solver):
        # Generate random data
        np.random.seed(0)
        m, n = 50, 4
        b = np.ones((m, 1))
        rand = np.random.randn(m - 2*n, n)
        A = np.vstack((rand, np.eye(n), np.eye(n) * -1))

        # Define the variable
        x = cp.Variable((n, 1))
        # set initial value for x
        objective = cp.Minimize(-cp.sum(cp.log(b - A @ x)))
        problem = cp.Problem(objective, [])
        # Solve the problem
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL

        checker = DerivativeChecker(problem)
        checker.run_and_assert()


    def test_socp(self, solver):
        # Define variables
        x = cp.Variable(3)
        y = cp.Variable()

        # Define objective function
        objective = cp.Minimize(3 * x[0] + 2 * x[1] + x[2])

        # Define constraints
        constraints = [
            cp.norm(x, 2) <= y,
            x[0] + x[1] + 3*x[2] >= 1.0,
            y <= 5
        ]

        # Create and solve the problem
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(objective.value, -13.548638814247532)
        assert np.allclose(x.value, [-3.87462191, -2.12978826, 2.33480343])
        assert np.allclose(y.value, 5)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_portfolio_socp(self, solver):
        np.random.seed(858)
        n = 100
        x = cp.Variable(n, name='x')
        mu = np.random.randn(n)
        Sigma = np.random.randn(n, n)
        Sigma = Sigma.T @ Sigma
        gamma = 0.1
        t = cp.Variable(name='t', bounds=[0, None])
        L = np.linalg.cholesky(Sigma, upper=False)

        objective = cp.Minimize(- mu.T @ x + gamma * t)
        constraints = [cp.norm(L.T @ x, 2) <= t,
                       cp.sum(x) == 1,
                       x >= 0]
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(problem.value, -1.93414338e+00)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_portfolio_socp_x_column_vector(self, solver):
        np.random.seed(858)
        n = 100
        x = cp.Variable((n, 1), name='x')
        mu = np.random.randn(n, 1)
        Sigma = np.random.randn(n, n)
        Sigma = Sigma.T @ Sigma
        gamma = 0.1
        t = cp.Variable(name='t', bounds=[0, None])
        L = np.linalg.cholesky(Sigma, upper=False)

        objective = cp.Minimize(-cp.sum(cp.multiply(mu, x)) + gamma * t)
        constraints = [cp.norm(L.T @ x, 2) <= t,
                       cp.sum(x) == 1,
                       x >= 0]
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(problem.value, -1.93414338e+00)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_localization(self, solver):
        np.random.seed(42)
        m = 10
        dim = 2
        x_true = np.array([2.0, -1.5])
        a = np.random.uniform(-5, 5, (m, dim))
        rho = np.linalg.norm(a - x_true, axis=1)  # no noise
        x = cp.Variable(2, name='x')
        t = cp.Variable(m, name='t')
        constraints = [t == cp.sqrt(cp.sum(cp.square(x - a), axis=1))]
        objective = cp.Minimize(cp.sum_squares(t - rho))
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(x.value, x_true)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_localization2(self, solver):
        np.random.seed(42)
        m = 10
        dim = 2
        x_true = np.array([2.0, -1.5])
        a = np.random.uniform(-5, 5, (m, dim))
        rho = np.linalg.norm(a - x_true, axis=1)  # no noise
        x = cp.Variable((1, 2), name='x')
        t = cp.Variable(m, name='t')
        constraints = [t == cp.sqrt(cp.sum(cp.square(x - a), axis=1))]
        objective = cp.Minimize(cp.sum_squares(t - rho))
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(x.value, x_true)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_circle_packing_formulation_one(self, solver):
        """Epigraph formulation."""
        rng = np.random.default_rng(5)
        n = 3
        radius = rng.uniform(1.0, 3.0, n)

        centers = cp.Variable((2, n), name='c')
        constraints = []
        for i in range(n - 1):
            for j in range(i + 1, n):
                constraints += [cp.sum(cp.square(centers[:, i] - centers[:, j])) >=
                                (radius[i] + radius[j]) ** 2]

        centers.value = rng.uniform(-5.0, 5.0, (2, n))
        t = cp.Variable()
        obj = cp.Minimize(t)
        constraints += [cp.max(cp.norm_inf(centers, axis=0) + radius) <= t]
        problem = cp.Problem(obj, constraints)
        problem.solve(solver=solver, nlp=True)

        true_sol = np.array([[1.73655994, -1.98685738, 2.57208783],
                             [1.99273311, -1.67415425, -2.57208783]])
        assert np.allclose(centers.value, true_sol)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    @pytest.mark.skipif('UNO' in INSTALLED_SOLVERS, reason='UNO finds a KKT point with worse obj')
    def test_circle_packing_formulation_two(self, solver):
        """Using norm_inf. This test revealed a very subtle bug in the unpacking of
        the ipopt solution. Some variables were mistakenly reordered. It was fixed
        in https://github.com/cvxgrp/cvxpy-ipopt/pull/82"""
        rng = np.random.default_rng(5)
        n = 3
        radius = rng.uniform(1.0, 3.0, n)

        centers = cp.Variable((2, n), name='c')
        constraints = []
        for i in range(n - 1):
            for j in range(i + 1, n):
                constraints += [cp.sum(cp.square(centers[:, i] - centers[:, j])) >=
                                (radius[i] + radius[j]) ** 2]

        centers.value = rng.uniform(-5.0, 5.0, (2, n))
        obj = cp.Minimize(cp.max(cp.norm_inf(centers, axis=0) + radius))
        prob = cp.Problem(obj, constraints)
        prob.solve(solver=solver, nlp=True)

        assert np.allclose(obj.value, 4.602738956101437)

        residuals = []
        for i in range(n - 1):
            for j in range(i + 1, n):
                dist_sq = np.linalg.norm(centers.value[:, i] - centers.value[:, j]) ** 2
                min_dist_sq = (radius[i] + radius[j]) ** 2
                residuals.append(dist_sq - min_dist_sq)

        assert(np.all(np.array(residuals) <= 1e-6))

        # Ipopt finds these centers, but Knitro rotates them (but finds the same
        # objective value)
        #true_sol = np.array([[1.73655994, -1.98685738, 2.57208783],
        #                     [1.99273311, -1.67415425, -2.57208783]])
        #assert np.allclose(centers.value, true_sol)

        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_circle_packing_formulation_three(self, solver):
        """Using max max abs."""
        rng = np.random.default_rng(5)
        n = 3
        radius = rng.uniform(1.0, 3.0, n)

        centers = cp.Variable((2, n), name='c')
        constraints = []
        for i in range(n - 1):
            for j in range(i + 1, n):
                constraints += [cp.sum(cp.square(centers[:, i] - centers[:, j])) >=
                                (radius[i] + radius[j]) ** 2]

        centers.value = rng.uniform(-5.0, 5.0, (2, n))
        obj = cp.Minimize(cp.max(cp.max(cp.abs(centers), axis=0) + radius))
        prob = cp.Problem(obj, constraints)
        prob.solve(solver=solver, nlp=True)

        # after the chain rule implementation, we find another configuration
        # with a minus sign
        true_sol = -np.array([[1.73655994, -1.98685738, 2.57208783],
                             [1.99273311, -1.67415425, -2.57208783]])
        assert np.allclose(centers.value, true_sol)

        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    @pytest.mark.skipif('UNO' in INSTALLED_SOLVERS, reason='UNO reaches iteration limit')
    def test_geo_mean(self, solver):
        x = cp.Variable(3, nonneg=True)
        geo_mean = cp.geo_mean(x)
        objective = cp.Maximize(geo_mean)
        constraints = [cp.sum(x) == 1]
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(x.value, np.array([1/3, 1/3, 1/3]))

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    @pytest.mark.skipif('UNO' in INSTALLED_SOLVERS, reason='UNO reaches iteration limit')
    def test_geo_mean2(self, solver):
        p = np.array([.07, .12, .23, .19, .39])
        x = cp.Variable(5, nonneg=True)
        prob = cp.Problem(cp.Maximize(cp.geo_mean(x, p)), [cp.sum(x) <= 1])
        prob.solve(solver=solver, nlp=True)
        x_true = p/sum(p)
        assert prob.status == cp.OPTIMAL
        assert np.allclose(x.value, x_true)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_div_composition(self, solver):
        x = cp.Variable(nonneg=True, bounds=[1, 5])
        prob = cp.Problem(cp.Maximize(cp.exp(1 / x)))
        prob.solve(solver=solver, nlp=True)
        x_true = 1.0
        assert prob.status == cp.OPTIMAL
        assert np.allclose(x.value, x_true)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_clnlbeam(self, solver):
        N = 1000
        h = 1 / N
        alpha = 350

        t = cp.Variable(N+1, bounds=[-1, 1])
        x = cp.Variable(N+1, bounds=[-0.05, 0.05])
        u = cp.Variable(N+1)
        u.value = np.zeros(N+1)
        control_terms = cp.multiply(0.5 * h, cp.power(u[1:], 2) + cp.power(u[:-1], 2))
        trigonometric_terms = cp.multiply(0.5 * alpha * h, cp.nlp.cos(t[1:]) + cp.nlp.cos(t[:-1]))
        objective_terms = cp.sum(control_terms + trigonometric_terms)

        objective = cp.Minimize(objective_terms)
        constraints = []
        position_constraints = (x[1:] - x[:-1] -
                                cp.multiply(0.5 * h, cp.nlp.sin(t[1:]) + cp.nlp.sin(t[:-1])) == 0)
        constraints.append(position_constraints)
        angle_constraint = (t[1:] - t[:-1] - 0.5 * h * (u[1:] + u[:-1]) == 0)
        constraints.append(angle_constraint)

        problem = cp.Problem(objective, constraints)
        problem.solve(solver=solver, nlp=True)
        assert problem.status == cp.OPTIMAL
        assert np.allclose(problem.value, 3.500e+02)

        # the derivative checker takes more than 10 seconds on this problem
        #checker = DerivativeChecker(problem)
        #checker.run_and_assert()



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import scipy.sparse as sp

from cvxpy.atoms.affine.binary_operators import multiply
from cvxpy.atoms.affine.reshape import reshape
from cvxpy.atoms.affine.transpose import swapaxes
from cvxpy.atoms.affine.upper_tri import upper_tri
from cvxpy.constraints.psd import PSD
from cvxpy.expressions.variable import Variable
from cvxpy.utilities.solver_context import SolverInfo


def lambda_max_canon(expr, args, solver_context: SolverInfo | None = None):
    A = args[0]
    n = A.shape[-1]
    batch_shape = A.shape[:-2]
    t = Variable(batch_shape)

    # Build t * I_n as (*batch, n, n):
    # reshape t to (*batch, 1, 1), elementwise multiply with eye(n)
    t_expanded = reshape(t, batch_shape + (1, 1), order='F')
    # Construct eye(n) as ND COO array with shape (*batch, n, n)
    diag = np.arange(n)
    batch_size = int(np.prod(batch_shape)) if batch_shape else 1
    data = np.ones(batch_size * n)
    rows = np.tile(diag, batch_size)
    cols = np.tile(diag, batch_size)
    if batch_shape:
        batch_flat = np.repeat(np.arange(batch_size), n)
        coords = np.unravel_index(batch_flat, batch_shape) + (rows, cols)
    else:
        coords = (rows, cols)
    eye_n = sp.coo_array((data, coords), shape=batch_shape + (n, n))
    t_eye = multiply(t_expanded, eye_n)  # (*batch, n, n)

    constr = [PSD(t_eye - A)]

    if not A.is_symmetric():
        constr.append(upper_tri(A) == upper_tri(swapaxes(A, -2, -1)))

    return t, constr



"""
Copyright 2026, the CVXPY Authors

Licensed under the Apache License, Version 2.0 (the "License");

you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""


import numpy as np

import cvxpy.settings as s
from cvxpy.constraints import SOC, ExpCone
from cvxpy.reductions.solution import Solution, failure_solution
from cvxpy.reductions.solvers import utilities
from cvxpy.reductions.solvers.conic_solvers.conic_solver import ConicSolver
from cvxpy.utilities.citations import CITATION_DICT


class PDCS(ConicSolver):
    """An interface for the PDCS solver.
    """
    # Solver capabilities.
    MIP_CAPABLE = False
    SUPPORTED_CONSTRAINTS = ConicSolver.SUPPORTED_CONSTRAINTS \
        + [SOC, ExpCone]
    STATUS_MAP = {
        "optimal" : s.OPTIMAL,
        "primal_infeasible_high_acc": s.INFEASIBLE,
        "dual_infeasible_high_acc": s.UNBOUNDED,
        "primal_infeasible_low_acc": s.INFEASIBLE_INACCURATE,
        "dual_infeasible_low_acc": s.UNBOUNDED_INACCURATE,
        "max_iter": s.USER_LIMIT,
        "time_limit": s.USER_LIMIT,
        "numerical_error": s.SOLVER_ERROR,
    }

    # Order of exponential cone arguments for solver.
    EXP_CONE_ORDER = [0, 1, 2]


    def name(self):
        """The name of the solver.
        """
        return 'PDCS'

    def import_solver(self) -> None:
        """Imports the solver.
        """
        import cupy  # noqa F401

    def supports_quad_obj(self) -> bool:
        """PDCS does not support quadratic objective with any combination
        of conic constraints.
        """
        return False

    def invert(self, solution, inverse_data):
        """Returns the solution to the original problem give the inverse_data.
        """
        attr = {}
        status = self.STATUS_MAP[str(solution.info.exit_status)]
        attr[s.SOLVE_TIME] = solution.info.time
        attr[s.NUM_ITERS] = solution.info.iter


        if status in s.SOLUTION_PRESENT:
            primal_val = solution.info.pObj
            opt_val = primal_val + inverse_data[s.OFFSET]
            primal_vars = {
                inverse_data[PDCS.VAR_ID]: np.array(solution.x.recovered_primal.primal_sol.x)
            }
            eq_dual_vars = utilities.get_dual_values(
                np.array(solution.y.recovered_dual.dual_sol.y[:inverse_data[ConicSolver.DIMS].zero]),
                utilities.extract_dual_value,
                inverse_data[PDCS.EQ_CONSTR]
            )
            ineq_dual_vars = utilities.get_dual_values(
                np.array(solution.y.recovered_dual.dual_sol.y[inverse_data[ConicSolver.DIMS].zero:]),
                utilities.extract_dual_value,
                inverse_data[PDCS.NEQ_CONSTR]
            )
            dual_vars = {}
            dual_vars.update(eq_dual_vars)
            dual_vars.update(ineq_dual_vars)
            return Solution(status, opt_val, primal_vars, dual_vars, attr)
        else:
            return failure_solution(status, attr)

    def solve_via_data(self, data, warm_start: bool, verbose: bool, solver_opts, solver_cache=None):
        """Returns the result of the call to the solver.

        Parameters
        ----------
        data : dict
            Data generated via an apply call.
        warm_start : Bool
            Whether to warm_start PDCS.
        verbose : Bool
            Control the verbosity.
        solver_opts : dict
            PDCS-specific solver options.
            Options can be passed in Python-native types (bool, int, float, str, etc.);
            booleans will be converted to "true"/"false" strings for Julia automatically.

        Returns
        -------
        The result returned by a call to PDCS_GPU.solve_with_solver().
        """

        # Helper to preprocess solver_opts so that bools and ints are stringified as Julia expects
        def normalize_opts(opts):
            norm_opts = {}
            for k, v in opts.items():
                if isinstance(v, bool):
                    norm_opts[k] = "true" if v else "false"
                elif isinstance(v, int) and k != "verbose":
                    # verbose can be left as int for logging level
                    norm_opts[k] = str(v)
                else:
                    norm_opts[k] = v
            return norm_opts

        # update default options first (as Python types)
        solver_opts.setdefault("abs_tol", 1e-6)
        solver_opts.setdefault("rel_tol", 1e-6)
        solver_opts.setdefault("logfile", "nothing")
        solver_opts.setdefault("time_limit_secs", 1000.0)
        solver_opts.setdefault("use_scaling", True)
        solver_opts.setdefault("rescaling_method", "ruiz_pock_chambolle")
        solver_opts.setdefault("use_adaptive_restart", True)
        solver_opts.setdefault("use_adaptive_step_size_weight", True)
        solver_opts.setdefault("use_resolving", True)
        solver_opts.setdefault("use_accelerated", False)
        solver_opts.setdefault("use_aggressive", True)
        solver_opts.setdefault("print_freq", 2000)
        solver_opts.setdefault("kkt_restart_freq", 2000)
        solver_opts.setdefault("duality_gap_restart_freq", 2000)
        solver_opts.setdefault("use_kkt_restart", False)
        solver_opts.setdefault("use_duality_gap_restart", True)
        solver_opts.setdefault("use_preconditioner", True)
        solver_opts.setdefault("method", "average")
        solver_opts.setdefault("julia_env", "placeholder")

        # handle verbose pythonically & normalize
        solver_opts["verbose"] = 1 if verbose else 0

        # Convert pythonic boolean/integer options to Julia-friendly strings where needed
        solver_opts = normalize_opts(solver_opts)
        import cupy
        from cupyx.scipy.sparse import csr_matrix as cucsr_matrix
        from juliacall import Main as jl

        if solver_opts["julia_env"] != "placeholder":
            jl.seval("using Pkg")
            jl.seval(f"Pkg.activate(\"{solver_opts['julia_env']}\")")
        # jl.seval(f"using Pkg")
        # jl.seval(f"Pkg.develop(path=\"./PDCS\")")
        # jl.seval(f"Pkg.resolve()")
        jl.seval('using CUDA, CUDA.CUSPARSE')
        jl.seval('using PDCS: PDCS_GPU, PDCS_CPU')
        A = data[s.A]
        b = data[s.B]
        q = data[s.C]
        if s.P in data:
            raise ValueError("PDCS does not support quadratic objective.")

        cones = data[ConicSolver.DIMS]
        cgpu = cupy.array(q)

        Ggpu = cucsr_matrix(A)
        bgpu = cupy.array(b)

        jl.c = jl.PDCS_GPU.cupy_to_cuvector(jl.Float64, int(cgpu.data.ptr), cgpu.size)

        jl.G = jl.PDCS_GPU.cupy_to_cucsrmat(
                jl.Float64, int(Ggpu.data.data.ptr), int(Ggpu.indices.data.ptr),
                int(Ggpu.indptr.data.ptr), *Ggpu.shape, Ggpu.nnz)
        jl.b = jl.PDCS_GPU.cupy_to_cuvector(jl.Float64, int(bgpu.data.ptr), bgpu.size)

        jl.soc = cones.soc
        jl.expG = cones.exp
        jl.m_zero = cones.zero
        jl.m_nonnegative = cones.nonneg
        assert len(cones.p3d) == 0, "PowCone3D is not supported"
        assert len(cones.pnd) == 0, "PowerCone is not supported"
        assert len(cones.psd) == 0, "PSD is not supported"
        jl.seval(f"""
        m, n = size(G)
        solver = PDCS_GPU.PDCS_GPU_Solver(
            n = n,
            m = m,
            nb = n,
            c = c,
            G = -G,
            h = -b,
            mGzero = m_zero,
            mGnonnegative = m_nonnegative,
            socG = Vector{{Integer}}(soc),
            rsocG = Vector{{Integer}}([]),
            expG = expG,
            dual_expG = 0,
            bl = CuArray(ones(n) * -Inf),
            bu = CuArray(ones(n) * Inf),
            soc_x = Vector{{Integer}}([]),
            rsoc_x = Vector{{Integer}}([]),
            exp_x = 0,
            dual_exp_x = 0,
            use_preconditioner = {solver_opts["use_preconditioner"]},
            method = :{solver_opts["method"]},
            print_freq = {solver_opts["print_freq"]},
            time_limit = {solver_opts["time_limit_secs"]},
            use_adaptive_restart = {solver_opts["use_adaptive_restart"]},
            use_adaptive_step_size_weight = {solver_opts["use_adaptive_step_size_weight"]},
            use_resolving = {solver_opts["use_resolving"]},
            use_accelerated = {solver_opts["use_accelerated"]},
            use_aggressive = {solver_opts["use_aggressive"]},
            verbose = {solver_opts["verbose"]},
            rel_tol = {solver_opts["rel_tol"]},
            abs_tol = {solver_opts["abs_tol"]},
            kkt_restart_freq = {solver_opts["kkt_restart_freq"]},
            duality_gap_restart_freq = {solver_opts["duality_gap_restart_freq"]},
            use_kkt_restart = {solver_opts["use_kkt_restart"]},
            use_duality_gap_restart = {solver_opts["use_duality_gap_restart"]},
            logfile_name = {solver_opts["logfile"]},
        )
        """)
        results = jl.PDCS_GPU.solve_with_solver(jl.solver)
        return results

    def cite(self, data):
        """Returns bibtex citation for the solver.

        Parameters
        ----------
        data : dict
            Data generated via an apply call.
        """
        return CITATION_DICT["PDCS"]



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import pytest

import cvxpy as cp
from cvxpy.reductions.solvers.defines import INSTALLED_SOLVERS
from cvxpy.tests.nlp_tests.derivative_checker import DerivativeChecker


@pytest.mark.skipif('IPOPT' not in INSTALLED_SOLVERS, reason='IPOPT is not installed.')
class TestScalarProblems():

    def test_exp(self):
        x = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.exp(x)), [x >= 4])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_entropy(self):
        x = cp.Variable()
        prob = cp.Problem(cp.Maximize(cp.entr(x)), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_KL(self):
       p = cp.Variable()
       q = cp.Variable()
       prob = cp.Problem(cp.Minimize(cp.kl_div(p, q)), [p >= 0.1, q >= 0.1, p + q == 2])
       prob.solve(nlp=True, solver=cp.IPOPT)
       assert prob.status == cp.OPTIMAL
       checker = DerivativeChecker(prob)
       checker.run_and_assert()

    def test_KL_matrix(self):
        Y = cp.Variable((3, 3))
        X = cp.Variable((3, 3))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.kl_div(X, Y))),
                        [X >= 0.1, Y >= 0.1, cp.sum(X) + cp.sum(Y) == 6])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_entropy_matrix(self):
        x = cp.Variable((3, 2))
        prob = cp.Problem(cp.Maximize(cp.sum(cp.entr(x))), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_logistic(self):
        x = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.logistic(x)), [x >= 0.4])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_logistic_matrix(self):
        x = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.logistic(x))), [x >= 0.4])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_power(self):
        x = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.power(x, 3)), [x >= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_power_matrix(self):
        x = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.power(x, 3))), [x >= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_power_fractional(self):
        x = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.power(x, 1.5)), [x >= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        x = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.power(x, 0.6))), [x >= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_power_fractional_matrix(self):
        x = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.power(x, 1.5))), [x >= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        x = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.power(x, 0.6))), [x >= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_scalar_trig(self):
        x = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.nlp.tan(x)), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        prob = cp.Problem(cp.Minimize(cp.nlp.sin(x)), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        prob = cp.Problem(cp.Minimize(cp.nlp.cos(x)), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_matrix_trig(self):
        x = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.nlp.tan(x))), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        prob = cp.Problem(cp.Minimize(cp.sum(cp.nlp.sin(x))), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        prob = cp.Problem(cp.Minimize(cp.sum(cp.nlp.cos(x))), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_matrix_hyperbolic(self):
        x = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.nlp.sinh(x))), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        prob = cp.Problem(cp.Minimize(cp.sum(cp.nlp.tanh(x))), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_scalar_hyperbolic(self):
        x = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.nlp.sinh(x)), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        prob = cp.Problem(cp.Minimize(cp.nlp.tanh(x)), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_xexp(self):
        x = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.xexp(x)), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        x = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.xexp(x))), [x >= 0.1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_scalar_quad_form(self):
        x = cp.Variable((1, ))
        P = np.array([[3]])
        prob = cp.Problem(cp.Minimize(cp.quad_form(x, P)), [x >= 1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_scalar_quad_over_lin(self):
        x = cp.Variable()
        y = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.quad_over_lin(x, y)), [x >= 1, y <= 1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_matrix_quad_over_lin(self):
        x = cp.Variable((3, 2))
        y = cp.Variable((1, ))
        prob = cp.Problem(cp.Minimize(cp.quad_over_lin(x, y)), [x >= 1, y <= 1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        y = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.quad_over_lin(x, y)), [x >= 1, y <= 1])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_rel_entr_both_scalar_variables(self):
        x = cp.Variable()
        y = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.rel_entr(x, y)),
                          [x >= 0.1, y >= 0.1, x <= 2, y <= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

        x = cp.Variable((1, ))
        y = cp.Variable((1, ))
        prob = cp.Problem(cp.Minimize(cp.rel_entr(x, y)),
                          [x >= 0.1, y >= 0.1, x <= 2, y <= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_rel_entr_matrix_variable_and_scalar_variable(self):
        x = cp.Variable((3, 2))
        y = cp.Variable()
        prob = cp.Problem(cp.Minimize(cp.sum(cp.rel_entr(x, y))),
                          [x >= 0.1, y >= 0.1, x <= 2, y <= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_rel_entr_scalar_variable_and_matrix_variable(self):
        x = cp.Variable()
        y = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.rel_entr(x, y))),
                          [x >= 0.1, y >= 0.1, x <= 2, y <= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_rel_entr_both_matrix_variables(self):
        x = cp.Variable((3, 2))
        y = cp.Variable((3, 2))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.rel_entr(x, y))),
                          [x >= 0.1, y >= 0.1, x <= 2, y <= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_rel_entr_both_vector_variables(self):
        x = cp.Variable((3, ))
        y = cp.Variable((3, ))
        prob = cp.Problem(cp.Minimize(cp.sum(cp.rel_entr(x, y))),
                          [x >= 0.1, y >= 0.1, x <= 2, y <= 2])
        prob.solve(nlp=True, solver=cp.IPOPT)
        assert prob.status == cp.OPTIMAL
        checker = DerivativeChecker(prob)
        checker.run_and_assert()



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from cvxpy.expressions.variable import Variable
from cvxpy.utilities.solver_context import SolverInfo


def cummax_canon(expr, args, solver_context: SolverInfo | None = None):
    """Cumulative max."""
    X = args[0]
    axis = expr.axis
    ndim = len(expr.shape)

    if expr.shape[axis] == 1:
        return X, []

    # Implicit O(n) definition:
    # Y_{k} = maximum(Y_{k-1}, X_k)
    Y = Variable(expr.shape)
    constr = [X <= Y]

    # Build slices for "all but last" and "all but first" along axis
    slice_prev = tuple(
        slice(None, -1) if i == axis else slice(None) for i in range(ndim)
    )
    slice_next = tuple(
        slice(1, None) if i == axis else slice(None) for i in range(ndim)
    )
    constr += [Y[slice_prev] <= Y[slice_next]]
    return Y, constr



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import pytest

import cvxpy as cp
from cvxpy.reductions.solvers.defines import INSTALLED_SOLVERS


@pytest.mark.skipif('IPOPT' not in INSTALLED_SOLVERS, reason='IPOPT is not installed.')
class TestBestOf():

    def test_circle_packing_best_of_one(self):
        np.random.seed(0)
        rng = np.random.default_rng(5)
        n = 5
        radius = rng.uniform(1.0, 3.0, n)
        centers = cp.Variable((n, 2), name='c')
        constraints = []
        for i in range(n - 1):
            constraints += [cp.sum((centers[i, :] - centers[i+1:, :]) ** 2, axis=1) >=
                            (radius[i] + radius[i+1:]) ** 2]
        obj = cp.Minimize(cp.max(cp.norm_inf(centers, axis=1) + radius))
        prob = cp.Problem(obj, constraints)
        centers.sample_bounds = [-5.0, 5.0]
        n_runs = 10
        prob.solve(nlp=True, verbose=True, derivative_test='none', best_of=n_runs)
        obj_best_of = obj.value
        best_centers = centers.value
        all_objs = prob.solver_stats.extra_stats['all_objs_from_best_of']

        assert len(all_objs) == n_runs
        manual_obj = np.max(np.linalg.norm(best_centers, ord=np.inf, axis=1) + radius)
        assert manual_obj == obj_best_of
        assert manual_obj == np.min(all_objs)

    def test_path_planning_best_of_two(self):
        # test that if sample bounds and the value of the variables are set,
        # then best_of still initializes randomly within the bounds
        np.random.seed(0)
        rng = np.random.default_rng(5)
        n = 5
        radius = rng.uniform(1.0, 3.0, n)
        centers = cp.Variable((n, 2), name='c')
        constraints = []
        for i in range(n - 1):
            constraints += [cp.sum((centers[i, :] - centers[i+1:, :]) ** 2, axis=1) >=
                            (radius[i] + radius[i+1:]) ** 2]
        obj = cp.Minimize(cp.max(cp.norm_inf(centers, axis=1) + radius))
        prob = cp.Problem(obj, constraints)

        centers.value = np.random.rand(n, 2)
        centers.sample_bounds = [-5.0, 5.0]
        n_runs = 10
        prob.solve(nlp=True, verbose=True, derivative_test='none', best_of=n_runs)
        obj_best_of = obj.value
        best_centers = centers.value
        all_objs = prob.solver_stats.extra_stats['all_objs_from_best_of']
        _, counts = np.unique(all_objs, return_counts=True)

        assert np.max(counts) == 1
        assert len(all_objs) == n_runs
        manual_obj = np.max(np.linalg.norm(best_centers, ord=np.inf, axis=1) + radius)
        assert manual_obj == obj_best_of
        assert manual_obj == np.min(all_objs)

    def test_path_planning_best_of_three(self):
        # test that no error is raised when best_of > 1 and all variables have finite bounds
        x = cp.Variable(bounds=[-5, 5])
        y = cp.Variable(bounds=[-3, 3])
        obj = cp.Minimize((x - 1) ** 2 + (y - 2) ** 2)
        prob = cp.Problem(obj)
        prob.solve(nlp=True, best_of=3)

        all_objs = prob.solver_stats.extra_stats['all_objs_from_best_of']
        assert len(all_objs) == 3

    def test_path_planning_best_of_four(self):
        # test that an error is raised it there is a variable with one
        # infinite bound and no sample_bounds when best_of > 1
        x = cp.Variable(bounds=[-5, 5])
        y = cp.Variable(bounds=[-3, None])
        obj = cp.Minimize((x - 1) ** 2 + (y - 2) ** 2)
        prob = cp.Problem(obj)

        # test that it raises an error
        with pytest.raises(ValueError):
            prob.solve(nlp=True, best_of=3)

    def test_path_planning_best_of_five(self):
        # test that no error is raised it there is a variable with
        # no bounds and no sample bounds, but it has been assigned
        # a value
        x = cp.Variable(bounds=[-5, 5])
        y = cp.Variable()
        y.value = 5
        obj = cp.Minimize((x - 1) ** 2 + (y - 2) ** 2)
        prob = cp.Problem(obj)
        prob.solve(nlp=True, best_of=3)
        all_objs = prob.solver_stats.extra_stats['all_objs_from_best_of']
        assert len(all_objs) == 3

    def test_best_of_infeasible_problem(self):
        # test that if the problem is infeasible, then best_of returns inf as the objective value
        x = cp.Variable(bounds=[-5, 5])
        y = cp.Variable(bounds=[-3, 3])
        constraints = [x + y == 10]
        obj = cp.Minimize((x - 1) ** 2 + (y - 2) ** 2)
        prob = cp.Problem(obj, constraints)
        prob.solve(nlp=True, best_of=20, verbose=True)
        assert prob.value == float("inf")

    def test_best_of_with_unbounded(self):
        # test that if the problem is unbounded, then best_of returns -inf as the objective value
        x = cp.Variable()
        x.sample_bounds = [-5, 5]
        obj = cp.Minimize(x)
        prob = cp.Problem(obj)
        prob.solve(nlp=True, best_of=20, verbose=True)
        assert prob.value == float("-inf")



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import numpy.linalg as LA
import pytest

import cvxpy as cp
from cvxpy.reductions.solvers.defines import INSTALLED_SOLVERS
from cvxpy.tests.nlp_tests.derivative_checker import DerivativeChecker

np.random.seed(0)

@pytest.fixture
def Sigma():
    return 1e-5 * np.array([
        [41.16, 22.03, 18.64, -4.74,  6.27, 10.1 , 14.52,  3.18],
        [22.03, 58.57, 32.92, -5.04,  4.02,  3.7 , 26.76,  2.17],
        [18.64, 32.92, 81.02,  0.53,  6.05,  2.02, 25.52,  1.56],
        [-4.74, -5.04,  0.53, 20.6 ,  2.52,  0.57,  0.2 ,  3.6 ],
        [6.27,  4.02,  6.05,  2.52, 10.13,  2.59,  4.32,  3.13],
        [10.1 ,  3.7 ,  2.02,  0.57,  2.59, 22.89,  3.97,  3.26],
        [14.52, 26.76, 25.52,  0.2 ,  4.32,  3.97, 29.91,  3.25],
        [3.18,  2.17,  1.56,  3.6 ,  3.13,  3.26,  3.25, 13.63]
    ])

@pytest.mark.skipif('IPOPT' not in INSTALLED_SOLVERS, reason='IPOPT is not installed.')
class TestRiskParity:
    def test_vanilla_risk_parity_formulation_one(self, Sigma):
        n = 8
        risk_target = (1 / n) * np.ones(n)

        w = cp.Variable((n,), nonneg=True, name='w')
        t = cp.Variable((n,), name='t')
        constraints = [cp.sum(w) == 1, t == Sigma @ w]

        term1 = cp.sum(cp.multiply(cp.square(w), cp.square(t))) / cp.quad_form(w, Sigma)
        term2 = (LA.norm(risk_target) ** 2) * cp.quad_form(w, Sigma)
        term3 = - 2 * cp.sum(cp.multiply(risk_target, cp.multiply(w, t)))
        obj = cp.Minimize(term1 + term2 + term3)
        problem = cp.Problem(obj, constraints)
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True, derivative_test='none')

        risk_contributions = w.value * (Sigma @ w.value)
        risk_contributions /= np.sum(risk_contributions)
        assert np.linalg.norm(risk_contributions - risk_target) < 1e-5

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    # we do not expand the objective, and use square roots
    def test_vanilla_risk_parity_formulation_two(self):
        pass

    # we expand the objective manually to get rid of the square root
    def test_group_risk_parity_formulation_one(self, Sigma):
        n = 8
        b = np.array([0.4, 0.6])

        w = cp.Variable((n, ), nonneg=True, name='w')
        t = cp.Variable((n, ), name='t')
        constraints = [cp.sum(w) == 1, t == Sigma @ w]
        w.value = np.ones(n) / n
        groups = [[0, 1, 5], [3, 4, 2, 6, 7]]

        term1 = 0
        term2 = 0
        term3 = 0

        for k, g in enumerate(groups):
            term1 += cp.square(cp.sum(cp.multiply(w[g], t[g]))) /  cp.quad_form(w, Sigma)
            term2 += (LA.norm(b[k]) ** 2) * cp.quad_form(w, Sigma)
            term3 += - 2 * b[k] * cp.sum(cp.multiply(w[g], t[g]))

        obj = cp.Minimize(term1 + term2 + term3)
        problem = cp.Problem(obj, constraints)
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True, derivative_test='none')
        risk_contributions = w.value * (Sigma @ w.value)
        risk_contributions /= np.sum(risk_contributions)
        risk_contributions = np.array([np.sum(risk_contributions[g]) for g in groups])
        assert np.linalg.norm(risk_contributions - b) < 1e-5

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    # other formulation
    def test_group_risk_parity_formulation_two(self, Sigma):
        n = 8
        b = np.array([0.4, 0.6])

        w = cp.Variable((n, ), nonneg=True, name='w')
        t = cp.Variable((n, ), name='t')
        constraints = [cp.sum(w) == 1, t == Sigma @ w]
        w.value = np.ones(n) / n
        groups = [[0, 1, 5], [3, 4, 2, 6, 7]]

        obj = 0
        for k, g in enumerate(groups):
            obj += cp.square(cp.sum(cp.multiply(w[g], t[g])) / cp.quad_form(w, Sigma) - b[k])

        problem = cp.Problem(cp.Minimize(obj), constraints)
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True, derivative_test='none')

        risk_contributions = w.value * (Sigma @ w.value)
        risk_contributions /= np.sum(risk_contributions)
        risk_contributions = np.array([np.sum(risk_contributions[g]) for g in groups])
        assert np.linalg.norm(risk_contributions - b) < 1e-5

        checker = DerivativeChecker(problem)
        checker.run_and_assert()



from cvxpy import settings as s
from cvxpy.reductions.reduction import Reduction


def _compose_id_map(step_maps):
    """Compose a sequence of ``{old_id: [new_id, ...]}`` mappings.

    Each step map comes from a reduction's ``var_id_map`` or
    ``param_id_map``.  The result is a single mapping from the
    outermost original IDs to lists of the innermost final IDs.
    """
    result = {}
    for step_map in step_maps:
        # Forward existing mappings through this step (1-to-many capable).
        result = {
            orig_id: [new_id for cur_id in cur_ids
                             for new_id in step_map.get(cur_id, [cur_id])]
            for orig_id, cur_ids in result.items()
        }
        # Add new mappings introduced by this reduction step.
        for orig_id in step_map.keys() - result.keys():
            result[orig_id] = list(step_map[orig_id])
    return result


class Chain(Reduction):
    """A logical grouping of multiple reductions into a single reduction.

    Attributes
    ----------
    reductions : list[Reduction]
        A list of reductions.
    """

    def __init__(self, problem=None, reductions=None) -> None:
        super(Chain, self).__init__(problem=problem)
        self.reductions = [] if reductions is None else reductions

    def __str__(self):
        return str(self.reductions)

    def __repr__(self) -> str:
        return "Chain(reductions=%s)" % repr(self.reductions)

    def get(self, reduction_type):
        for reduction in self.reductions:
            if isinstance(reduction, reduction_type):
                return reduction
        raise KeyError

    def accepts(self, problem) -> bool:
        """A problem is accepted if the sequence of reductions is valid.

        In particular, the i-th reduction must accept the output of the i-1th
        reduction, with the first reduction (self.reductions[0])
        in the sequence taking as input the supplied problem.

        Parameters
        ----------
        problem : Problem
            The problem to check.

        Returns
        -------
        bool
            True if the chain can be applied, False otherwise.
        """

        for r in self.reductions:
            if not r.accepts(problem):
                return False
            problem, _ = r.apply(problem)
        return True

    def apply(self, problem, verbose: bool = False):
        """Applies the chain to a problem and returns an equivalent problem.

        Parameters
        ----------
        problem : Problem
            The problem to which the chain will be applied.
        verbose : bool, optional
            Whehter to print verbose output.

        Returns
        -------
        Problem or dict
            The problem yielded by applying the reductions in sequence,
            starting at self.reductions[0].
        list
            The inverse data yielded by each of the reductions.
        """
        inverse_data = []
        for r in self.reductions:
            if verbose:
                s.LOGGER.info('Applying reduction %s', type(r).__name__)
            problem, inv = r.apply(problem)
            inverse_data.append(inv)
        return problem, inverse_data

    def compose_var_id_map(self):
        """Compose variable ID mappings across all reductions.

        Returns a single ``{orig_var_id: [final_var_id, ...]}`` dict that
        accounts for every variable replacement in the chain.  If
        reduction *i* maps ``A → [A']`` and reduction *j* (j > i) maps
        ``A' → [A'']``, the result contains ``A → [A'']``.  If a step
        maps ``A' → [A'', A''']``, the result expands to
        ``A → [A'', A''']``.

        Returns
        -------
        dict
            Maps original variable IDs to lists of their final (innermost)
            IDs.
        """
        return _compose_id_map(r.var_id_map for r in self.reductions)

    def compose_param_id_map(self):
        """Compose parameter ID mappings across all reductions.

        Returns a single ``{orig_param_id: [final_param_id, ...]}`` dict
        that accounts for every parameter replacement in the chain.

        Returns
        -------
        dict
            Maps original parameter IDs to lists of their final (innermost)
            IDs.
        """
        return _compose_id_map(r.param_id_map for r in self.reductions)

    def compose_constr_id_map(self):
        """Compose constraint ID mappings across all reductions.

        Returns a single ``{orig_constr_id: final_constr_id}`` dict tracing
        each original constraint ID through every reduction step.  Each
        reduction exposes its mapping via the ``cons_id_map`` property, which
        is populated during ``apply()``.

        Returns
        -------
        dict
            Maps original constraint IDs to their final (innermost) IDs.
        """
        result = {}
        for r in self.reductions:
            step_map = r.cons_id_map
            if step_map:
                result = {orig: step_map.get(cur, cur) for orig, cur in result.items()}
                for orig_id, new_id in step_map.items():
                    if orig_id not in result:
                        result[orig_id] = new_id
        return result

    def invert(self, solution, inverse_data):
        """Returns a solution to the original problem given the inverse_data.
        """
        for r, inv in reversed(list(zip(self.reductions, inverse_data))):
            solution = r.invert(solution, inv)
        return solution
